// Animales (C) Agosto de 2.000
// El ordenador intentar adivinar en que animal esta pensando el jugador.
// Ejemplo de aplicacin de arboles binarios, y como almacenarlos y
// recuperarlos de fichero.
// Es adems una forma bsica de sistema experto. Por medio de preguntas
// el programa es capaz de diagnosticar averias, o en este caso adivinar
// el animal en que el jugador esta pensando.
// Versin C++, con clases

#include <stdio.h>
#include <string.h>

// Declaracion de Nodo, si los punteros a otros nodos son nulos, se trata de una
// respuesta, en caso contrario se trata de una pregunta.
class Nodo {
	public:
    Nodo(): Si(0), No(0), Texto(0) {}	// Constructor por defecto
    Nodo(char *);								// Constructor con un texto definido
    ~Nodo();									// Destructor

    char *Texto;								// Campo de respuesta o pregunta
    Nodo *Si;									// Nodo si la respuesta es SI
    Nodo *No;                          // Nodo si la respuesta es NO
};

// Declaracion de Arbol binario, base de datos de conocimiento
class Arbol {
	public:
    Arbol(): Raiz() {}						// Constructor, crea un nodo raiz nulo
    Arbol(char *);							// Constructor, crea un arbol leyendolo
    												// desde un archivo de disco
    ~Arbol();									// Destructor
	 void VerArbol();                   // Muestra el arbol en pantalla
	 void InsertarAnimal(Nodo *, bool, char *, bool, char *);
    												// Insercion de nuevo animal

    Nodo *Raiz;                        // Nodo raiz, entrada en la base de datos

   private:
    Nodo *aux;                         // Nodo auxiliar para funciones recursivas
	 char Nombre[256];                  // Nombre de fichero de la base de datos
	 Nodo *LeeArbol(FILE *);				// Lee el arbol desde un fichero
    void EscribeArbol(FILE *, Nodo*);  // Escribe el arbol a fichero
													// Funciones recursivas:
    void Borrar(Nodo *);               // Borra el arbol a partir de un Nodo
    void VerSubArbol(Nodo *);				// Muestra el arbol a partir de un nodo
    int n;                             // Variable auxiliar para indentar la
    												// salida de ver arbol
};

// Constructor de Nodo
Nodo::Nodo(char *texto)
{
	Texto = new char[strlen(texto)+1];	// reserva memoria para el texto
   strcpy(Texto, texto);               // Copia el texto a memoria dinamica
   Si = 0;                             // Por defecto es un nodo "hoja"
   No = 0;
}

// Destructir de Nodo
Nodo::~Nodo()
{
	if(Texto) delete[] Texto;				// Si existe el texto, libera la memoria
}

// Constructor de Arbol desde fichero
Arbol::Arbol(char *nombre)
{
	FILE *fe;									// Fichero de entrada

   strcpy(Nombre, nombre);             // Guarda el nombre de fichero para
   												// usarlo al salvar el arbol
   fe = fopen(Nombre, "r");            // Abre el fichero en modo lectura

   if(!fe)                             // si no existe el fichero...
      Raiz = new Nodo("Caballo");	   // Crea un nodo hoja para el Caballo
   else
   {
   	Raiz = LeeArbol(fe);             // si existe lee recursivamente...
   	fclose(fe);                      // cierra el fichero cuando termina
   }
}

// Destructor de Arbol
Arbol::~Arbol()
{
   FILE *fs;									// Fichero de salida

   fs = fopen(Nombre, "w");				// Abre o crea el fichero de salida

   EscribeArbol(fs, Raiz);					// Escribe en fs el arbol recursivamente
   fclose(fs);									// Cierra el fichero

	Borrar(Raiz);                       // Destruye todos los nodos
}

// Funcion recursina de lectura de arbol
Nodo *Arbol::LeeArbol(FILE *fe)
{
   char Linea[258];							// Auxiliar para lectura de lineas
   Nodo *A;										// Auxiliar para aadir nodos

   fgets(Linea, 257, fe);					// Lee una linea desde el fichero fe
   												// Eliminamos retornos de linea
  	while(Linea[strlen(Linea) - 1] < ' ' && Linea[strlen(Linea) - 1] > 0)
   	Linea[strlen(Linea) - 1] = 0;

   A = new Nodo(&Linea[1]);				// Crea un nuevo nodo, nos saltamos el
   												// primer caracter, que indica el tipo

   if(Linea[0] == 'P')						// es un nodo pregunta
   {
	   A->Si = LeeArbol(fe);				// Leemos rama de "si"
      A->No = LeeArbol(fe);       		// Leemos rama de "no"
   }
   return A;                           // Valor de retorno
}

//	Funcion recusiva de escritura de un arbol a fichero
void Arbol::EscribeArbol(FILE *fs, Nodo *A)
{
	if(A->Si)									// Si es un nodo pregunta
   {
   	fprintf(fs, "P%s\n", A->Texto);	// Escribe 'P'+ pregunta en fs
      EscribeArbol(fs, A->Si);			// Escribe rama de Si
      EscribeArbol(fs, A->No);         // Escribe rama de No
   }
   else											// Es un nodo respuesta
   	fprintf(fs, "R%s\n", A->Texto);	// Escribe 'R'+ repuesta en fs
}

// Funcion recursiva de borrado de nodos
void Arbol::Borrar(Nodo *n)
{
   if(n->Si) Borrar(n->Si);				// Borrar rama n->Si
   if(n->No) Borrar(n->No);   			// Borrar rama n->No
   delete n;                  			// Destruir nodo n
}

// Funcion de mostrar arbol en pantalla
void Arbol::VerArbol()
{
   n = 0;                              // n = 0, empezar con indentado nulo
	VerSubArbol(Raiz);						// Funcion recursiva de ver arbol
}

// Funcion recursiva de mostrar arbol en pantalla
void Arbol::VerSubArbol(Nodo *A)
{
	int i;            						// variable auxiliar

	for(i = 0; i < n; i++) putchar(' ');// Indentado
	n += 3;                             // Aumentar indentado
   printf(">%s\n", A->Texto);          // Muestra texto
   if(A->Si)
	{
   	printf("S ");							// respuesta afirmativa
      VerSubArbol(A->Si);              // Muestra texto
   }
   if(A->No)
   {
   	printf("N ");							// respuesta negativa
      VerSubArbol(A->No);              // Muestra texto
   }
	n -= 3;										// Disminuir indentado
}

// Inserta un animal nuevo, implica insertar una nueva pregunta
void Arbol::InsertarAnimal(Nodo *Anterior, bool RespuestaAnterior,
	char* Pregunta, bool Resp, char *AnimalNuevo)
{
  	/* Ahora hemos de aadir el nodo nuevo
   Estado anterior:
   	+-----------+
      | Pregunta  + <-- Anterior
   	+-----------+
     	  |       |
        |       |  <-- RespuestaAnterior
                |
   		+-----------+
   P-->  | Animal 1  +  <-- Ultimo animal preguntado, respuesta negativa
         +-----------+
        NO |       | SI
           |       |
          NULL    NULL

   Estado final puede ser:
   a) Si Resp = SI
   +-----------+
   | Pregunta  + <-- Anterior
   +-----------+
     |       |
     |       |  <-- RespuestaAnterior
             |
   		+----------------+
   Q-->  | Pregunta nueva +
         +----------------+
        NO |       | SI
           |       |
   +-----------+ +--------------+
   | Animal 1  + | Nuevo Animal +
   +-----------+ +--------------+
   b) Si Resp = NO
   +-----------+
   | Pregunta  + <-- Anterior
   +-----------+
     |       |
     |       |  <-- RespuestaAnterior
             |
   		+----------------+
   Q-->  | Pregunta nueva +
         +----------------+
        SI |       | NO
           |       |
   +-----------+ +--------------+
   | Animal 1  + | Nuevo Animal +
   +-----------+ +--------------+
   */

   Nodo *P, *Q;					// variables auxiliares

   if(RespuestaAnterior)		// P debe apuntar al animal 1
   	P = Anterior->Si;
   else
   	P = Anterior->No;

   Q = new Nodo(Pregunta); 	// Creamos un nodo nuevo para nueva pregunta
	if(RespuestaAnterior)		// Enlazamos la nueva pregunta en su sitio
     	Anterior->Si = Q;
   else
    	Anterior->No = Q;

   if(Resp)							// Enlazamos el animal anterior en su sitio
     	Q->No = P;					// Caso a) Rama de Q->No es el animal 1
   else
     	Q->Si = P;					// Caso b) Rama de Q->Si es el animal 1

   P = new Nodo(AnimalNuevo); // Creamos un nuevo nodo pana nuevo animal

   if(Resp)							// Enlazamos el animal nuevo en su sitio
     	Q->Si = P;					// Caso a) Rama de Q->Si es el animal nuevo
   else
    	Q->No = P;         		// Caso b) Rama de Q->No es el animal nuevo
}

// Prototipos
bool Respuesta();             // Funcion de lectura de respuestas

int main()
{
	Arbol *Datos = new Arbol("Data.pan");	// Base de datos en forma de arbol

   Nodo *Q, *P, *Anterior;						// Auxiliares para navegacion
   bool Resp, RespuestaAnterior;				// Auxiliares para respuestas
   char Nuevo[256], Pregunta[256];			// Textos para aadir nodos

   do // Bucle de navegacion, mientras el jugador quiera seguir jugando...
   {
   	/* Muestra mensaje y espera a que el jugador este preparado */
      printf("Piensa en un animal y yo intentar adivinarlo.\n");
      printf("Pulsa una tecla para comenzar.\n");
      Respuesta();			// En realidad no nos improta la respuesta...

      Anterior = NULL;		// Aun no hemos pasado por ningun nodo
      P = Datos->Raiz;		// P apunta al nodo raiz, P sera siempre el nodo
      							// actual

      do							// Bucle, se permanece en este bucle mientras
      							// no se alcance una hoja, (es decir una respuesta)
      {
			Q = P;				// Guardamos un puntero a P, para preguntar en el
         						// do...while si hemos alcanzado una hoja
         if(P->Si)         // Si existen nodos en cualquiera de las ramas
         						// se trata de una pregunta, si no es una respuesta
            printf("%s?\n", P->Texto);	// Muestra la pregunta al jugador
         else
            printf("Se trata de: %s\n", P->Texto); // Muestra la respuesta

			Resp = Respuesta(); // Lee la respuesta del jugador

         if(P->Si)			// Si aun hay nodos, moverse a la rama adecuada
         {
         	Anterior = P;  // Guarda un puntero al nodo que sera el anterior
            RespuestaAnterior = Resp; // Guarda nuestra respuesta, para saber
            								  // en que direccion hemos ido
            if(Resp) P = P->Si; // cambiamos de nodo
            else P = P->No;     // dependiendo de la respuesta
         }
      }
      while(Q->Si);			// do...while, permanece en el buicle miestras haya
      							// preguntas por hacer

      if(Resp)					// Si la ultima respuesta fue Si, el programa ha
      							// adivinado el animal
         printf("Lo he adivinado!!!\n"); // Se congratula...
      else						// En caso contrario, es que no conoce el animal
      {
         printf("No conozco este animal, pero si me respondes a unas\n");
         printf("preguntas lo incluire en mi lista.\n");
         do						// Este bucle asegura que no se aadiran nodos
         						// con errores
         {
            printf("\nDe que animal se trata?\n");
            fgets(Nuevo, 255, stdin);	// Lee el nombre del animal
			   while(Nuevo[strlen(Nuevo) - 1] < ' ')
            	Nuevo[strlen(Nuevo) - 1] = 0; // Elimina retorno de linea
            printf("Dame una pregunta que sirva para distinguir un/a\n");
            printf("%s\nde un/a\n%s\n", P->Texto, Nuevo);
            printf("a la que se pueda contestar si o no:\n");
            fgets(Pregunta, 255, stdin);// Lee la pregunta para distinguir el
         						// nuevo animal del ultimo por el que se pregunto
			   while(Pregunta[strlen(Pregunta) - 1] < ' ')
            	Pregunta[strlen(Pregunta) - 1] = 0; // Elimina retorno de linea
            printf("Que respuesta se ha de dar a esta pregunta:\n");
            printf("%s?\n", Pregunta);
            printf("Para obtener: %s\n", Nuevo);
            printf("como respuesta: Si o No:\n");
            Resp = Respuesta();			// Lee la respuesta adecuada a la
            									// pregunta leida

				// Muestra al jugador los datos leidos para que verifique si
            // son correctos
            printf("Veamos si lo he entendido bien:\n");
            printf("A la pregunta: %s?\n", Pregunta);
            if(Resp)
            {
               printf("Si se responde SI se trata de un %s\n", Nuevo);
               printf("y si se responde NO de un %s\n", P->Texto);
            }
            else
            {
               printf("Si se responde SI se trata de un %s\n", P->Texto);
               printf("y si se responde NO de un %s\n", Nuevo);
            }
            printf("Es correcto?");
         }
         while(!Respuesta());	// Repite el proceso hasta que el jugador de el
         							// visto bueno

      	// Inserta la nueva pregunta y el nuevo animal adecuadamente:
         Datos->InsertarAnimal(Anterior, RespuestaAnterior,
         		Pregunta, Resp, Nuevo);
      }

      // Pregunta al jugador si quiere seguir jugando
      printf("Deseas terminar el juego y salir? (S/N)");
   }
   while(!Respuesta());	// Permanece jugando mientras se responda que si

   Datos->VerArbol();	// Solo mientras se depura, ver base de datos
   delete Datos;			// Destruye arbol, se salva en disco antes.
	return 0;				// Volvemos al SO
}

// Lee respuestas del tipo Si/No, retorna con true para Si y con false para No
bool Respuesta()
{
   char Respuesta[10];  // Variable para recoger respuestas del tipo Si/No

   // Lee la respuesta del jugador
	fgets(Respuesta, 9, stdin);	// Sobran caracteres, pero mejor prevenir...

   return Respuesta[0] == 'S' || Respuesta[0] == 's'; // Valor de retorno
   				// Se considera respuesta afirmativa 'S' o 's'
}


